@@ -6,6 +6,7 @@ from typing import Any
 
 from markdown_it import MarkdownIt
 from markdown_it.token import Token
+from mdformat import codepoints
 from mdformat.renderer import DEFAULT_RENDERER_FUNCS, LOGGER, RenderTreeNode
 from mdformat.renderer.typing import RendererFunc
 
@@ -94,7 +95,7 @@ def _render_heading(
     env: MutableMapping,
 ) -> str:
     if not _toc_enabled(env):
-        DEFAULT_RENDERER_FUNCS["heading"](node, renderer_funcs, options, env)
+        return DEFAULT_RENDERER_FUNCS["heading"](node, renderer_funcs, options, env)
 
     heading_idx = env["mdformat-toc"]["rendered_headings"]
     heading = env["mdformat-toc"]["headings"].headings[heading_idx]
@@ -109,17 +110,20 @@ def _render_html_block(
     env: MutableMapping,
 ) -> str:
     if not _toc_enabled(env) or not is_toc_start_node(node):
-        DEFAULT_RENDERER_FUNCS["html_block"](node, renderer_funcs, options, env)
+        return DEFAULT_RENDERER_FUNCS["html_block"](node, renderer_funcs, options, env)
 
     opts = env["mdformat-toc"]["opts"]
     text = f"<!-- mdformat-toc start {opts} -->\n\n"
 
-    text += _render_toc(
+    toc = _render_toc(
         env["mdformat-toc"]["headings"],
         minlevel=opts.minlevel,
         maxlevel=opts.maxlevel,
     )
-    text += "\n<!-- mdformat-toc end -->"
+    if toc:
+        text += toc + "\n"
+
+    text += "<!-- mdformat-toc end -->"
 
     return text
 
@@ -146,7 +150,8 @@ def _render_toc(
 
     for heading in heading_tree.headings:
         indentation = "  " * heading_tree.get_indentation_level(heading)
-        toc += f"{indentation}- [{heading.text}](<#{heading.slug}>)\n"
+        uri = _maybe_add_link_brackets("#" + heading.slug)
+        toc += f"{indentation}- [{heading.text}]({uri})\n"
 
     return toc
 
@@ -260,3 +265,12 @@ def _ensure_anchors_in_place(heading_tokens: Sequence[Token]) -> None:
         Token("html_inline", "", 0, content="</a>"),
     ]
     inline_root.children += link_tokens
+
+
+def _maybe_add_link_brackets(link: str) -> str:
+    """Surround URI with brackets if required by the CommonMark spec."""
+    if not link or (
+        codepoints.ASCII_CTRL | codepoints.ASCII_WHITESPACE | {"(", ")"}
+    ).intersection(link):
+        return "<" + link + ">"
+    return link
